﻿using System;
using System.Collections.Generic;
using System.Diagnostics.CodeAnalysis;
using System.Globalization;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Security;
using vm.Aspects.Diagnostics;

namespace vm.Aspects
{
    public static class Extensions
    {
        /// <summary>
        /// Dumps the specified <paramref name="value"/> to a text writer.
        /// </summary>
        /// <param name="value">The object to dump.</param>
        /// <param name="writer">The text writer to dump to.</param>
        /// <param name="indentLevel">The initial indent level.</param>
        /// <param name="dumpMetadata">
        /// Optional metadata class to use to extract the dump attributes from. If not specified, the dump metadata will be sought in
        /// a <see cref="MetadataTypeAttribute"/> attribute applied to the <paramref name="value"/>'s class. And if that is not found - from the 
        /// attributes applied within the class itself.
        /// </param>
        /// <param name="classDumpAttribute">
        /// An explicit dump attribute to be applied at a class level. If not specified the <see cref="MetadataTypeAttribute"/> attribute applied to 
        /// <paramref name="obj"/>'s class and if that is not specified - <see cref="MetadataTypeAttribute.Default"/> will be assumed.
        /// </param>
        /// <returns>
        /// The object.
        /// </returns>
        [SuppressMessage("Microsoft.Design", "CA1031:DoNotCatchGeneralExceptionTypes", Justification="Any exceptions here are most likely bugs which should not kill the process.")]
        [SecuritySafeCritical]
        public static object DumpText(
            this object value,
            TextWriter writer,
            int indentLevel = 0,
            Type dumpMetadata = null,
            DumpAttribute dumpAttribute = null)
        {
            if (writer == null)
                throw new ArgumentNullException("writer");

            try
            {
                var dumper = new ObjectTextDumper(writer, indentLevel);

                dumper.Dump(value, dumpMetadata, dumpAttribute);
            }
            catch (Exception x)
            {
                writer.WriteLine(
                    "\n\nATTENTION:\nThe TextDumper threw an exception:\n{0}",
                    x.ToString());
            }

            return value;
        }

        /// <summary>
        /// Dumps the <paramref name="value"/> to a string.
        /// </summary>
        /// <param name="value">The object to be dumped.</param>
        /// <param name="indentLevel">The initial indent level.</param>
        /// <returns>
        /// The text dump of the object.
        /// </returns>
        [SecuritySafeCritical]
        public static string DumpString(
            this object value,
            int indentLevel = 0,
            Type dumpMetadata = null,
            DumpAttribute dumpAttribute = null)
        {
            using (var writer = new StringWriter(CultureInfo.InvariantCulture))
            {
                value.DumpText(writer, indentLevel, dumpMetadata, dumpAttribute);
                return writer.GetStringBuilder().ToString();
            }
        }

        /// <summary>
        /// Determines whether the specified type is basic: primitive, enum, decimal, string, Guid, Uri, DateTime, TimeSpan, DateTimeOffset, IntPtr, 
        /// UIntPtr.
        /// </summary>
        /// <param name="type">The type to be tested.</param>
        /// <returns>
        ///   <c>true</c> if the specified type is one of the basic types; otherwise, <c>false</c>.
        /// </returns>
        internal static bool IsBasicType(
            this Type type)
        {
            if (type == null)
                throw new ArgumentNullException("type");

            return
                type.IsEnum              ||
                type.IsPrimitive         ||
                type == typeof(DBNull)   ||
                type == typeof(decimal)  ||
                type == typeof(string)   ||
                type == typeof(Guid)     || 
                type == typeof(Uri)      ||
                type == typeof(DateTime) ||
                type == typeof(TimeSpan) ||
                type == typeof(IntPtr)   ||
                type == typeof(UIntPtr)  ||
                type == typeof(DateTimeOffset);
        }

        /// <summary>
        /// Retrieves a custom attribute of a specified type that is applied to a specified member.
        /// </summary>
        /// <typeparam name="T">The type of attribute to search for.</typeparam>
        /// <param name="attributeProvider">The member to inspect.</param>
        /// <param name="inherit"><see langword="true" /> to inspect the ancestors of element; otherwise, <see langword="false" /> (the default).</param>
        /// <returns>A custom attribute that matches <typeparamref name="T"/>, or <see langword="null" /> if no such attribute is found.</returns>
        /// <exception cref="System.ArgumentNullException">Thrown if <paramref name="attributeProvider"/> is <see langword="null"/>.</exception>
        /// <exception cref="System.InvalidOperationException">Thrown if more than one attribute of the specified type <typeparamref name="T"/> is found.</exception>
        public static T GetCustomAttribute<T>(this ICustomAttributeProvider attributeProvider, bool inherit = false) where T : class
        {
            if (attributeProvider == null)
                throw new ArgumentNullException("attributeProvider");

            if (!attributeProvider.IsDefined(typeof(T), inherit))
                return null;

            return (T)attributeProvider.GetCustomAttributes(typeof(T), inherit).Single();
        }

        /// <summary>
        /// Retrieves all custom attributes of a specified type that are applied to a specified member.
        /// </summary>
        /// <typeparam name="T">The type of attribute to search for.</typeparam>
        /// <param name="attributeProvider">The member to inspect.</param>
        /// <param name="inherit"><see langword="true" /> to inspect the ancestors of element; otherwise, <see langword="false" /> (the default).</param>
        /// <returns>A sequence of custom attributes that match <typeparamref name="T"/> (possibly empty).</returns>
        /// <exception cref="System.ArgumentNullException">Thrown if <paramref name="attributeProvider"/> is <see langword="null"/>.</exception>
        public static IEnumerable<T> GetCustomAttributes<T>(this ICustomAttributeProvider attributeProvider, bool inherit = false) where T : class
        {
            if (attributeProvider == null)
                throw new ArgumentNullException("attributeProvider");

            if (!attributeProvider.IsDefined(typeof(T), inherit))
                return new T[0];

            return attributeProvider.GetCustomAttributes(typeof(T), inherit).OfType<T>();
        }

        /// <summary>
        /// Determines whether the specified <see cref="PropertyInfo"/> object represents a virtual property.
        /// </summary>
        /// <param name="pi">The <see cref="PropertyInfo"/> object.</param>
        /// <returns><c>true</c> if the specified <see cref="PropertyInfo"/> object represents a virtual property; otherwise, <c>false</c>.</returns>
        /// <exception cref="System.ArgumentNullException">Thrown if <paramref name="pi"/> is <see langword="null"/>.</exception>
        public static bool? IsVirtual(this PropertyInfo pi)
        {
            if (pi == null)
                throw new ArgumentNullException("pi");

            bool? isVirtual = null;

            foreach (var accessor in pi.GetAccessors())
                if (isVirtual.HasValue)
                {
                    if (isVirtual.Value != accessor.IsVirtual)
                        return null;
                }
                else
                    isVirtual = accessor.IsVirtual;

            return isVirtual;
        }
    }
}
